! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************

!==================================================================================
!
! This module contains the first simple, coherent presecription of surface fossil magnetic fields
! in massive star evolutionary models.
!
! The new routines included below are part of the PhD Thesis of 
! Z. Keszthelyi 2019, Queen's Univesity, Kingston, Canada. 
!  
! The first application was presented by Petit, Keszthelyi, MacInnis, 
! et al. 2017, MNRAS, 540, 466, 1052 in the case of non-rotating models. 
! Since then, three published papers, using the Geneva code, 
! adopted a similar, coherent picture (Georgy, et al. 2017, A&A, 599, 5; 
! Song, et al. (2018) A&A, 609, 3; 
! and Keszthelyi, Meynet, Georgy, et al. 2019, MNRAS, 485, 5843 - Paper I). 
! 
! The implementation adopted in this module is described in detial 
! by Keszthelyi, Meynet, Shultz, et al. 2019 - Paper II. 
! 
! The module is applicable to model known magnetic stars that possess surface fossil magnetic fields.
! Those fields form a large-scale magnetoshpere (see, e.g., 
! Donati & Landstreet 2009, ARA&A, 47, 333, for a review). 
!
! Following the works of ud-Doula & Owocki 2002, ApJ, 576, 413;
! ud-Doula, Owocki, Townsend 2008, MNRAS, 385, 97;
! and ud-Doula, Owocki, Townsend 2009, MNRAS, 392, 1022,
! we adopt a formalism, which is applicable in one-dimensional codes.
!
! We consider three ingredients to model surface fossil magnetic fields: 
!
! Mass-loss quenching reduces the effective mass-loss rate of the star (other_wind routine).
! Magnetic braking reduces the angular momentum of the star (other_torque routine).
! The fossil field evolves by time, following the frozen flux theorem (Alfven, 1942). 
!
! We caution that there are several uncertainties and 
! this first approach needs to be further improved.
! 
! Zsolt Keszthelyi et al. 
!
!==================================================================================
module run_star_extras

  use const_lib
  use const_def      
  use star_private_def  ! THIS ALWAYS HAS TO BE COPIED MANUALLY AFTER NEW INSTALLATION! it is in star/make and has to be in ../include
  use star_lib
  use star_def
  use chem_def, only: ih1, ihe4, ic12, ic13, in14, io16

  implicit none

      ! Module variables: 
      !==========================
      real(dp) :: R_init  ! initial radius, calc. actual value in mag routine
      real(dp) :: Bp      ! polar magnetic field strength 
      real(dp) :: B0      ! Assumed initial field (add in inlist, star_job) -- uncomment this if want to use initial flux
      real(dp) :: vesc    ! escape velocity (calculated for electron scattering opacity)
      real(dp) :: vinf    ! terminal velocity
      real(dp) :: etastar ! wind magnetic confinement parameter
      real(dp) :: Ra      ! Alfven radius
      real(dp) :: Rc      ! closure radius  
      real(dp) :: f_B     ! escaping mass fraction allowed by the magnetosphere
      real(dp) :: logf_B
      real(dp) :: f_rot     ! rotational enhancement following Maeder & Meynet (2000)
      real(dp) :: Mdot_orig ! original mass-loss rate (unaffected by the magnetosphere) 
      real(dp) :: w_vink
      real(dp) :: w_graefener
      real(dp) :: w_vanloon
      real(dp) :: Jbrake    ! dJ / dt from ud-Doula et al.  
      real(dp) :: J_surf    ! AM in the near-surface of the star 
      real(dp) :: scale     ! to (uniformly) distribute AM loss
      real(dp) :: core_h1_init  ! initial core H content
      integer, parameter :: UNIFORM_TORQUE = 1    ! the entire star slows its rotation
      integer, parameter :: SURFACE_TORQUE = 2    ! the surface layers slow their rotation
      integer, save  :: torque_method = 0 ! Method to evaluate torque
      !==========================

 contains
      
  subroutine extras_controls(id, ierr)

    integer, intent(in) :: id
    integer, intent(out) :: ierr
    
    type (star_info), pointer :: s

    ierr = 0

    call star_ptr(id, s, ierr)
    if (ierr /= 0) return

    s% extras_startup => extras_startup
    s% extras_check_model => extras_check_model
    s% extras_finish_step => extras_finish_step
    s% extras_after_evolve => extras_after_evolve
    s% how_many_extra_history_columns => how_many_extra_history_columns
    s% data_for_extra_history_columns => data_for_extra_history_columns
    s% how_many_extra_profile_columns => how_many_extra_profile_columns
    s% data_for_extra_profile_columns => data_for_extra_profile_columns  

    s% other_torque => MAG_braking          ! torque_routine
    s% other_wind => MAG_WIND               ! wind_routine
    ! Once you have set the function pointers you want,
    ! then uncomment this (or set it in your star_job inlist)
    ! to disable the printed warning message,

    s% job% warn_run_star_extras =.false.       
    
    ! ====================================================
    ! Set magnetic parameters in the inlist file (star_job) 
    ! B0 is the initial polar magnetic field strength in gauss
    ! The torque method will apply the torque for either the entire star (UNIFORM)
    ! or the surface AM reservior (SURFACE)
    ! ====================================================

    if (s%job%extras_lrpar >= 1) then
       B0 = s%job%extras_rpar(1)
    else
       stop 'Initial field strength is not set using extras_rpar(1)'
    endif

    if (s%job%extras_lcpar >= 1) then

       select case (s%job%extras_cpar(1))
       case ('UNIFORM_TORQUE')
          torque_method = UNIFORM_TORQUE
       case ('SURFACE_TORQUE')
          torque_method = SURFACE_TORQUE
       case default
          stop 'Torque method is not set.'
       end select

    else
       stop 'Torque method is not set using extras_cpar(1)'
    endif
            
  end subroutine extras_controls

 ! ====================================================     
 
  integer function extras_startup(id, restart, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         integer :: h1
         ierr = 0
         call star_ptr(id, s, ierr)
         call get_star_ptr(id,s,ierr)
         extras_startup = 0

         ! Usual startup stuff
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
         
         ! Store the initial H core mass fraction
         ! This will be used to calculate the ZAMS radius 
         ! for magnetic flux conservation
         !

          h1 = s% net_iso(ih1)
          core_h1_init = s% xa(h1,s% nz)

          R_init = 0.0d0 !!  R_init is set to zero for the first model. 
                         !! it is then calculated in the magnetic routine.

          WRITE (*,*) ':::--- Initial surface magnetic field strength in G ---:::'
          WRITE (*,*)  B0
          WRITE (*,*) '======================================================='
          WRITE (*,*) 'Torque method selected (uniform or surface)'
          WRITE (*,*)  torque_method
          WRITE (*,*) '======================================================='

      end function extras_startup

  ! ====================================================
  
  integer function extras_check_model(id, id_extra)

    integer, intent(in) :: id, id_extra
    type (star_info), pointer :: s
    integer :: ierr, i
    logical :: is_ne_biggest
    extras_check_model = keep_going
   ! ierr = 0
   ! call star_ptr(id, s, ierr)
   ! if (ierr /= 0) return

  end function extras_check_model

  !****
integer function how_many_extra_history_columns(id, id_extra)
   integer, intent(in) :: id, id_extra
   integer :: ierr
   type (star_info), pointer :: s
   ierr = 0
   call star_ptr(id, s, ierr)
   if (ierr /= 0) return
   how_many_extra_history_columns = 15     ! don't forget to synchronize this.
end function how_many_extra_history_columns
      
  !****
      
  integer function how_many_extra_profile_columns(id, id_extra)
    use star_def, only: star_info
    integer, intent(in) :: id, id_extra
    integer :: ierr
    type (star_info), pointer :: s
    ierr = 0
    call star_ptr(id, s, ierr)
    if (ierr /= 0) return
    how_many_extra_profile_columns = 1   ! don't forget to synchronize this.
  end function how_many_extra_profile_columns

  !****

  integer function extras_finish_step(id, id_extra)
    integer, intent(in) :: id, id_extra
    integer :: ierr
    integer :: h1 
    real(dp) :: core_h1
    type (star_info), pointer :: s
    real(dp), parameter :: F_LIMIT = 0.25_dp

    ierr = 0
    call star_ptr(id, s, ierr)
    if (ierr /= 0) return
    
    extras_finish_step = keep_going
    call store_extra_info(s)
    
    ! ====================================================
    ! Time-step control implemented by R. Townsend
    !
    ! Limit dt_next based on the torquing factor:
    ! This avoids possible issues with exhausting or exceeding angular momentum in a given layer
    ! and therefore reduces the time-step if necessary. 
    !
    ! This might cause the simulation to become very computational time-expensive! 
    ! ====================================================
    
    if (scale > F_LIMIT) then
       s%dt_next = s%dt_next/(scale/F_LIMIT)
       WRITE(*,*) '====----------====='
       WRITE(*,*) 'A better next time step is:'
       WRITE(*,*) s%dt_next/secyer
       WRITE(*,*) '====----------====='
    endif

   !! ======= Profile for models at any custom time ==========
   ! if (s% star_age >= 0.999d6 .AND. s% star_age < 1.001d6) then
   !  call star_write_profile_info(s% id, 'models/prof1Myr.data', s% id, ierr) 
   !  WRITE(*,*) 'A structure profile is saved.'
   ! end if
   
   h1 = s% net_iso(ih1)
   core_h1 = s% xa(h1,s% nz)

   !! ======= Profile for models at any custom X_c ==========
   ! if (core_h1 <  0.7201 .AND. core_h1 > 0.7199) then
   !  call star_write_profile_info(s% id, 'models/prof_ZAMS.data', s% id, ierr) 
   !  WRITE(*,*) 'A structure profile is saved.'
   ! end if

    ! see extras_check_model for information about custom termination codes
    ! by default, indicate where (in the code) MESA terminated
    if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step

  end function extras_finish_step

  !****
      
  subroutine extras_after_evolve(id, id_extra, ierr)
    integer, intent(in) :: id, id_extra
    integer, intent(out) :: ierr
    type (star_info), pointer :: s
    ierr = 0
    call star_ptr(id, s, ierr)
    if (ierr /= 0) return
  end subroutine extras_after_evolve
  
  !****

  subroutine alloc_extra_info(s)
    integer, parameter :: extra_info_alloc = 1
    type (star_info), pointer :: s
    call move_extra_info(s,extra_info_alloc)
  end subroutine alloc_extra_info

  !****
      
  subroutine unpack_extra_info(s)
    integer, parameter :: extra_info_get = 2
    type (star_info), pointer :: s
    call move_extra_info(s,extra_info_get)
  end subroutine unpack_extra_info

  !****
      
  subroutine store_extra_info(s)
    integer, parameter :: extra_info_put = 3
    type (star_info), pointer :: s
    call move_extra_info(s,extra_info_put)
  end subroutine store_extra_info

  !****
      
  subroutine move_extra_info(s,op)
    integer, parameter :: extra_info_alloc = 1
    integer, parameter :: extra_info_get = 2
    integer, parameter :: extra_info_put = 3
    type (star_info), pointer :: s
    integer, intent(in) :: op
    
    integer :: i, j, num_ints, num_dbls, ierr
         
    i = 0
    ! call move_int or move_flg    
    num_ints = i
         
    i = 0
    ! call move_dbl       
    
    num_dbls = i

    if (op /= extra_info_alloc) return
    if (num_ints == 0 .and. num_dbls == 0) return

    ierr = 0
    call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
    if (ierr /= 0) then
       write(*,*) 'failed in star_alloc_extras'
       write(*,*) 'alloc_extras num_ints', num_ints
       write(*,*) 'alloc_extras num_dbls', num_dbls
       stop 1
    end if

  contains

    subroutine move_dbl(dbl)
      real(dp) :: dbl
      i = i+1
      select case (op)
      case (extra_info_get)
         dbl = s% extra_work(i)
      case (extra_info_put)
         s% extra_work(i) = dbl
      end select
    end subroutine move_dbl

    subroutine move_int(int)
      integer :: int
      i = i+1
      select case (op)
      case (extra_info_get)
         int = s% extra_iwork(i)
      case (extra_info_put)
         s% extra_iwork(i) = int
      end select
    end subroutine move_int

    subroutine move_flg(flg)
      logical :: flg
      i = i+1
      select case (op)
      case (extra_info_get)
         flg = (s% extra_iwork(i) /= 0)
      case (extra_info_put)
         if (flg) then
            s% extra_iwork(i) = 1
         else
            s% extra_iwork(i) = 0
         end if
      end select
    end subroutine move_flg

  end subroutine move_extra_info

!=========================================================
!!!! ******** Main additions from here *******************


! Note that since we need vinf, we calculate it explicitly from 
! the Eddington parameter for pure electron scattering.
! We also adopt the bi-stability, that is the behaviour of vinf/vesc. and Mdot at a given Teff.
! This should be carefully checked to be consistent with the wind description.

! === Terminal velocity using Gamma for electron scattering (Kudritzki et al. 1986)=== 
! === Rotational enhancement following Maeder & Meynet (2000) === 
! 
! gammae routine based on the implementation of 
! Keszthelyi, Puls, Wade, 2017, A&A 584, A4 (KPW2017)
!

subroutine gammae(id,ierr)
  type (star_info), pointer :: s
  integer, intent(in) :: id
  integer, intent(out) :: ierr
  integer :: h1, he4, nz
  real(dp) :: alfa1, alfa2, L1, M1, T1, R1 ! alfa1,2 used for interpolation only
  real(dp) :: I_he, YHe, sigma_e, Gamma_e  ! Gamma_e is the Eddington factor for electron scattering  
  real(dp) :: vinf1, vinf2, vinf3, logvinf 
  real(dp) :: surface_h1, surface_he4   
  real(dp) :: nom, denom, pow
  real(dp) :: f_rot1, f_rot2, f_rot3
  real(dp) :: X, Y, Z  ! H, He and metal mass fractions
  real(dp), parameter :: alpha = 0.66d0    ! force multiplier parameter representative of O stars - 
                                           ! otherwise Teff dependence needs to be included

! --- Values related to the bi-stability --- check consistency with wind routine!

! The calculations are partitioned into three Teff regimes. 
  real(dp), parameter :: Teff_jump1 = 20.d3   ! Revised bi-stability jump temperature. See Petrov, et al. (2016), 
                                              ! KPW2017, and Vink (2018).
! FIRST bi-stability jump temperature (from observations)
  real(dp), parameter :: Teff_jump2 = 10.d3  
! SECOND bi-stability jump temperature (from observations) 
  real(dp), parameter :: dT1 = 1.d3   ! adopted for consistency with Vink routine
! Interpolation width of the first bi-stability jump. 
! Note that a small dT results in a steep change
! in Mdot while a larger dT yields a more gradual change. 
  real(dp), parameter :: dT2 = 1.d3   ! second jump 
! Interpolation width of the second bi-stability jump. 

! --- Values related to terminal wind velocity ---
! adopted from Kudritzki & Puls (2000), ARA&A, 38, 613; and Markova & Puls (2008), A&A, 478, 823
  real(dp), parameter :: fvinf1 = 2.6d0  ! ratio of terminal to escape velocity before Teff_jump_1
  real(dp), parameter :: fvinf2 = 1.3d0  ! ratio of terminal to escape velocity after Teff_jump_1 and before Teff_jump_2
  real(dp), parameter :: fvinf3 = 0.7d0  ! ratio of terminal to escape velocity after Teff_jump_2

  call get_star_ptr(id,s,ierr)

  L1 = s% photosphere_L         ! cgs
  M1 = s% star_mass * Msun      ! cgs
  T1 = s% Teff                  ! cgs
  R1 = s% photosphere_r * Rsun  ! cgs

!--- elements ---- 
  nz = s% nz
  h1 = s% net_iso(ih1)
  he4 = s% net_iso(ihe4)
  surface_h1 = s% xa(h1,1)
  surface_he4 = s% xa(he4,1)
  X = surface_h1            ! mass fraction of H
  Y = surface_he4           ! mass fraction of He
  Z = 1.d0 - (X + Y)        ! surface metallicity 

  YHe = Y/X / 4.d0  
! surface Helium content by number, 
! n_He/n_H; approximation: factor 4 should be m_He/m_H

! === PATH 1 === 
  ! --- vinf behaviour --- check for consistency with wind routine 
 if (T1 .ge. Teff_jump1 - dT1) then 
    I_he = 2.d0  ! the number of free electrons per He nuclei (hot regime)
    sigma_e = 0.398d0 * (1.d0 + YHe * I_he) / (1.d0 + 4.d0 * YHe)  ! Eq. 66 by Kudritzki et al. (1989) A&A, 219, 205 
    Gamma_e =  sigma_e * L1/(M1 * pi4 * standard_cgrav * clight) 
! escape velocity corrected for electron scattering 
    vesc = sqrt((1.d0 - Gamma_e) * 2.d0 * standard_cgrav * M1 / R1 )
! Eq. 15 KPW2017
    vinf1 = vesc * fvinf1
    vinf = vinf1
! rotational enhancement 
    nom = 1.d0 - Gamma_e
    denom = 1.d0 - (4.d0/9.d0 * pow_cr(s%v_rot_avg_surf / s%v_crit_avg_surf, 2.0d0) ) - Gamma_e  ! Note v_crit definition is different! 
    pow = (1.d0 / alpha) - 1.d0
    f_rot1 = pow_cr(nom, pow) / pow_cr(denom, pow)  
    f_rot = f_rot1
endif
! === PATH 2 === 
 if (T1 .le. Teff_jump1 + dT1 .and. T1 .ge. Teff_jump2 - dT2) then 
    I_he = 1.d0     ! cooler side of jump, Helium recombined to HeII 
    sigma_e = 0.398d0 * (1.d0 + YHe * I_he) / (1.d0 + 4.d0 * YHe)  
    Gamma_e =  sigma_e * L1/(M1 * pi4 * standard_cgrav * clight) 
    vesc = sqrt(2.d0 * standard_cgrav * M1 / R1 * (1.d0-Gamma_e))
    vinf2 = vesc * fvinf2
    vinf = vinf2
! rotational enhancement 
    nom = 1.d0 - Gamma_e
    denom = 1.d0 - (4.d0/9.d0 * pow_cr(s%v_rot_avg_surf / s%v_crit_avg_surf, 2.0d0) ) - Gamma_e  
    pow = (1.d0 / alpha) - 1.d0
    f_rot2 = pow_cr(nom, pow) / pow_cr(denom, pow) 
    f_rot = f_rot2
 endif
! --- INTERPOLATION for the FIRST bi-stability jump ---  
 if (Teff_jump1 + dT1 > T1 .and. T1 > Teff_jump1 - dT1) then
    alfa1 = (T1 - (Teff_jump1 - dT1)) / (2.d0 * dT1)
    vinf = alfa1 * vinf1 + (1.d0 - alfa1) * vinf2
    f_rot = alfa1 * f_rot1 + (1.d0 - alfa1) * f_rot2
 endif
! === PATH 3 ===
 if (T1 .le. Teff_jump2 + dT2) then
    sigma_e = 0.398d0 * 1.d0 / (1.d0 + 4.d0 * YHe) ! since I_he = 0 for T1 < 10 kK. 
    Gamma_e =  sigma_e * L1/(M1 * pi4 * standard_cgrav * clight) 
    vesc = sqrt(2.d0 * standard_cgrav * M1 / R1 * (1.d0-Gamma_e))   
    vinf3 = vesc * fvinf3
    vinf = vinf3
! rotational enhancement 
    nom = 1.d0 - Gamma_e
    denom = 1.d0 - (4.d0/9.d0 * pow_cr(s%v_rot_avg_surf / s%v_crit_avg_surf, 2.0d0) ) - Gamma_e  
    pow = (1.d0 / alpha) - 1.d0
    f_rot3 = pow_cr(nom, pow) / pow_cr(denom, pow) 
    f_rot = f_rot3
 endif
! --- INTERPOLATION for the SECOND bi-stability jump ---    
 if (Teff_jump2 + dT2 > T1 .and. T1 > Teff_jump2 - dT2) then
    alfa2 = (T1 - (Teff_jump2 - dT2)) / (2.d0 * dT2)  
    vinf = alfa2 * vinf2 + (1.d0 - alfa2) * vinf3
    f_rot = alfa2 * f_rot2 + (1.d0 - alfa2) * f_rot3
 endif


! To avoid numerical problems, if f_rot is too small (or rotation is very slow), 
! just ignore the rotational enhahncement on the mass-loss rates. 

if (f_rot .le. 1.d-4 .OR. s%v_rot_avg_surf .le. 1.d0) then   
 f_rot = 1.d0
endif 

end subroutine gammae

! ========================================================================
! === Calculation of MAGNETIC parameters including field evolution === 
!
! Routine used by Petit, Keszthelyi, MacInnis, et al. 2017, MNRAS, 466, 1052,
! and Keszthelyi, Meynet, Shultz, et al. 2019 (Paper II)
!
! ========================================================================

subroutine magnetic(id,ierr)
  type (star_info), pointer :: s
  integer, intent(in) :: id
  integer, intent(out) :: ierr
  integer :: h1 
  real(dp) :: L1, M1, T1, R1, core_h1  
  real(dp) :: logB0, logBp, logetastar
  real(dp) :: logvesc, logvinf

!====  Initial magnetic flux can also be set.
!      Set only if B0 is not set. 
!
!  real(dp), parameter :: F0 = 10**28    ! initial magnetic flux in units of G cm2
!
!====

 call gammae(id,ierr)  ! call in gammae for vinf routine  
 call get_star_ptr(id,s,ierr)

! basic parameters 
  L1 = s% photosphere_L ! cgs
  M1 = s% star_mass * Msun   ! cgs
  T1 = s% Teff      ! cgs
  R1 = s% photosphere_r * Rsun  ! cgs

  h1 = s% net_iso(ih1)
  core_h1 = s% xa(h1,s% nz)

! ====================================
! Setting the ZAMS radius here
! We adopt this ZAMS definition: some hydrogen is already fused in the core
! to generate the stellar luminosity
!
  if (core_h1 <= (core_h1_init * 0.997d0) .AND. R_init <= 0.d0) then 
     R_init = s% photosphere_r * 6.9598d10
  end if 
! ====================================

 if (.not. s% doing_relax) then       ! don't do calc. for relaxation phase 

! If the initial flux F0 is set instead of B0, then B0 needs to be calculated.
! F0 and R_init are in cgs, pi4 = 4. * pi
!
! B0 = F0/(pi4 * R_init * R_init)    ! this equation considers the unsigned flux.
!

! =====================================
!  ===  MAGNETIC FLUX CONSERVATION ===
! =====================================
!
  logBp = log10_cr(B0) + 2.d0 * log10_cr(R_init) - 2.d0 * log10_cr(R1)  ! cgs 
  Bp = exp10_cr(logBp)
!
! ====================================

! === Equatorial magnetic confinement parameter ===                 ! in cgs
  logetastar = 2.d0 * logBp + 2.d0 * log10_cr(R1)    &
               - log10_cr(Mdot_orig * Msun/secyer)   &
               - log10_cr(vinf) - log10_cr(4.0d0)
  etastar = exp10_cr(logetastar)

! === Alfve'n and closure radii ===
  Ra = 1.0d0 + pow_cr(etastar + 0.25d0, 0.25d0) - pow_cr(0.25d0, 0.25d0) ! this is: R_A / R_star
  Rc = 1.0d0 + 0.7d0 * (Ra - 1.0d0)                                      ! this is: R_c / R_star

! === Mass-loss quenching parameter ===    ! used to multiply original Mdot in wind routine
  f_B = 1.0d0 - sqrt(1.0d0 - 1.0d0/Rc)      
  logf_B = log10_cr(f_B)                   

! For rotating models, this scaling will need to consider
! an updated version, using Eq. 22 from ud-Doula et al. (2009).
! The update is only needed if the star spends a significant time 
! with a centrifugal magnetoshpere, and even then the correction factor is small. 

 endif  ! for not doing relax. 

end subroutine magnetic

! ====================================
!      === MAGNETIC BRAKING === 
! ====================================

subroutine MAG_braking(id, ierr)
        integer, intent(in) :: id
        real(dp) :: R1, M1, T1, Mdot, logvesc 
        real(dp) :: logBp, logvinf, logetastar, Omega, Jtotnew, komega, qpart
        integer, intent(out)      :: ierr
        type (star_info), pointer :: s
        integer                   :: k_max
        integer                   :: k, nz  
         real(dp)                 :: J_lost, J_ex
        ierr = 0
        call star_ptr(id, s, ierr)
        if (ierr /= 0) return

 if (s% doing_relax) then    
 write(*,*) 'still relaxing a bit - you should do the same'
 write(*,*) ''
 end if 
!-------------------------------
 if (.not. s% doing_relax .AND. R_init> 0._dp .AND. s%mstar_dot < 0._dp) then

  R1 = s% photosphere_r * Rsun  ! cgs               
  M1 = s% photosphere_m * Msun  ! cgs

  Mdot =  Mdot_orig * Msun / secyer          ! in cgs ! this is the original mass-loss rate for B=0.
  Omega = s% omega(1)   ! s% omega_avg_surf  ! in history this is 'surf_avg_omega', why???

  J_lost = s%angular_momentum_removed   ! by the mass loss

  ! === Total angular momentum loss due to magnetic fields === 
  ! This is dJ(field+gas)/dt from ud Doula, Owocki, Townsend (2009)
  ! Ra is in R_star units, so now it is converted to cgs. 
  ! The negative sign will follow later to extract this quantity.  

  Jbrake = (2.0d0 / 3.0d0) * Mdot * Omega * (Ra*R1)**2.d0    

  J_ex = max(Jbrake * s% dt - J_lost,0._dp)   ! avoid double counting the AM removed by mass-loss.

  if (J_ex < 0._dp) then
   WRITE (*,*) ':::::::::::::'
   WRITE (*,*) 'J ex cannot be negative.'
   WRITE (*,*) Mdot, Omega, s% v_rot_avg_surf, Ra, Jbrake, J_ex
   ierr = -1 
   return
  end if

  if (J_ex >= 0._dp) then  

    k_max = s%k_const_mass   ! number of constant mass zone

! ==================================================================
! Two torque methods are introduced because the exact 
! penetration depth of the fossil field remains unknown. 
!
! UNIFORM will remove AM from the entire star, SURFACE will only brake the surface rotation. 
! This is a simplifying assumption because we are not manipulating the angular momentum transport, 
! only the loss. 
! 
    select case (torque_method)  ! set in inlist. 

    case (UNIFORM_TORQUE)        ! convert J_ex to extra_j throughout the entire star.
  
  ! Jbrake converted to specific AM via scale:
  ! the same fraction of specific AM is removed from layer to layer but not the same amount.
  
    scale = J_ex / s% total_angular_momentum
     
      if (scale < 1._dp .AND. scale > 1.d-15) then
        do k = 1,s%nz  
          s%extra_jdot(k) = - s% j_rot(k) *  scale / s% dt    ! negative sign added here: this is AM loss. 
        end do
       else 
        do k = 1, s%nz
         s% extra_jdot(k) = 0._dp
        end do          
       end if

     case (SURFACE_TORQUE)        ! convert J_ex to extra_j in the near-surface of the star.

     J_surf = DOT_PRODUCT(s%dm_bar(1:k_max), s%j_rot(1:k_max))
     scale = J_ex  / J_surf

      if (scale < 1._dp .AND. scale > 1.d-15) then
        do k = 1,k_max  
          s%extra_jdot(k) = - s% j_rot(k) *  scale / s% dt    ! negative sign added here: this is AM loss.
        end do
       else 
        do k = 1,k_max
         s% extra_jdot(k) = 0._dp
        end do          
       end if

      case default
       stop 'Invalid torque method'

      end select  ! method
    end if        ! Jbrake not zero 
   end if         ! not doing relaxation
  
end subroutine MAG_braking


!=============================================================
! === 'Default' WIND routines: Vink, Graefener, de Jager, van Loon === 
!       (add here any other wind routines as required)
!=============================================================

!--- MESA's default Vink routine (updated following KPW2017)  ---

subroutine Vink(id, L_phot, M_phot, R_phot, T_phot, w, ierr)  

  type (star_info), pointer :: s
  real(dp), intent(in) :: L_phot, M_phot, R_phot, T_phot ! photospheric values (cgs)
  integer, intent(in) :: id
  integer, intent(out) :: ierr
  integer :: h1, he4, nz
  real(dp) :: X, Y, Z, L1, M1, R1, T1, surface_h1, surface_he4
  logical, parameter :: dbg = .false.
  real(dp), intent(out) :: w
  real(dp) :: alfa, w1a, w2a, w1, w2, w3, Teff_jump, logMdot, dT, vinf_div_vesc, logMdot_orig
  real(dp), parameter :: Zsolar = 0.019d0   ! Vink et al. use Z = 0.019 as solar
  
  call get_star_ptr(id,s,ierr)  

  nz = s% nz
  h1 = s% net_iso(ih1)
  he4 = s% net_iso(ihe4)

  surface_h1 = s% xa(h1,1)
  surface_he4 = s% xa(he4,1)

  L1 = L_phot   ! cgs
  M1 = M_phot   ! cgs
  T1 = T_phot   ! cgs
  R1 = s% photosphere_r * Rsun  ! cgs
  !sqrt(L1/(pi*crad*clight*T1*T1*T1*T1))      ! crad = 4 * sigma / c ...
                                              ! hence: pi*4*sigma*c/c
  X = surface_h1
  Y = surface_he4       ! mass fraction
  Z = 1 - (X + Y)       ! surface metallicity 

            ! needs to be consistent with vinf calc. in gammae subrutine !
            ! alfa = 1 for hot side, = 0 for cool side
            if (T1 > 21.d3) then   ! Note the updated bi-stability jump temperature ! 
               alfa = 1
            else if (T1 < 19.d3) then
               alfa = 0
            else ! previously used the Vink et al 2001, eqns 14 and 15 to set "jump" temperature
                 ! 1d3*(61.2d0 + 2.59d0*(-13.636d0 + 0.889d0*log10_cr(Z/Zsolar)))
               Teff_jump = 20d3   ! Updated value, see KPW2017
               dT = 1.d3          ! changed, see KPW2017
               if (T1 > Teff_jump + dT) then
                  alfa = 1
               else if (T1 < Teff_jump - dT) then
                  alfa = 0
               else
                  alfa = (T1 - (Teff_jump - dT)) / (2*dT)
               end if
            end if

            if (alfa > 0) then ! eval hot side wind (eqn 24)
               vinf_div_vesc = 2.6d0 ! this is the hot side galactic value
               vinf_div_vesc = vinf_div_vesc*pow_cr(Z/Zsolar,0.13d0) ! corrected for Z
               logMdot = &
                  - 6.697d0 &
                  + 2.194d0*log10_cr(L1/Lsun/1d5) &
                  - 1.313d0*log10_cr(M1/Msun/30) &
                  - 1.226d0*log10_cr(vinf_div_vesc/2d0) &
                  + 0.933d0*log10_cr(T1/4d4) &
                  - 10.92d0*pow2(log10_cr(T1/4d4)) &
                  + 0.85d0*log10_cr(Z/Zsolar)             
               w1a = exp10_cr(logMdot)               ! in Msun/yr
            else
               w1a=0
            end if
               
            if (alfa < 1) then ! eval cool side wind (eqn 25)
               vinf_div_vesc = 1.3d0 ! this is the cool side galactic value
               vinf_div_vesc = vinf_div_vesc*pow_cr(Z/Zsolar,0.13d0) ! corrected for Z
               logMdot = &
                  - 6.688d0 &
                  + 2.210d0*log10_cr(L1/Lsun/1d5) &
                  - 1.339d0*log10_cr(M1/Msun/30) &
                  - 1.601d0*log10_cr(vinf_div_vesc/2d0) &
                  + 1.07d0*log10_cr(T1/2d4) &
                  + 0.85d0*log10_cr(Z/Zsolar) 
                  w2a = exp10_cr(logMdot)             ! in Msun/yr
            else
               w2a=0
            end if     
            w = alfa*w1a + (1 - alfa)*w2a    
            w_vink = w  

       !    if (dbg) write(*,*) 'wind', w_vink
    
end subroutine Vink

!--- MESA's default Graefener routine ---

subroutine Graefener(id, L_phot, M_phot, R_phot, T_phot, w, ierr)
         ! Grafener, G. & Hamann, W.-R. 2008, A&A 482, 945
         ! routine contributed by Nilou Afsari
  type (star_info), pointer :: s
  real(dp), intent(in) :: L_phot, M_phot, R_phot, T_phot ! photospheric values (cgs)
  integer ::  h1, he4, nz
  real(dp) :: X, Y, Z, surface_h1, surface_he4, L1, M1, T1
  real(dp), intent(inout) :: w
  real(dp) :: w1, logMdot, gamma_edd, xsurf, beta, gammazero, lgZ
  integer, intent(in) :: id
  integer, intent(out) :: ierr
  real(dp), parameter :: Zsolar = 0.019d0  

  call get_star_ptr(id,s,ierr)  ! pointer to get data

  L1 = L_phot 
  M1 = M_phot
  T1 = T_phot
 
  nz = s% nz
  h1 = s% net_iso(ih1)
  he4 = s% net_iso(ihe4)

  surface_h1 = s% xa(h1,1)
  surface_he4 = s% xa(he4,1)

  X = surface_h1
  Y = surface_he4     
  Z = 1 - (X + Y) 

           xsurf = surface_h1
           gamma_edd = exp10_cr(-4.813d0)*(1+xsurf)*(L1/Lsun)*(Msun/M1)
           lgZ = log10_cr(Z/Zsolar)
           beta = 1.727d0 + 0.250d0*lgZ
           gammazero = 0.326d0 - 0.301d0*lgZ - 0.045d0*lgZ*lgZ
           logMdot = &
               + 10.046d0 &
               + beta*log10_cr(gamma_edd - gammazero) &
               - 3.5d0*log10_cr(T1) &
               + 0.42d0*log10_cr(L1/Lsun) &
               - 0.45d0*xsurf
            w = exp10_cr(logMdot)
            w_graefener = w    
       ! if (dbg) write(*,*) 'grafener wind', w
end subroutine Graefener


!subroutine deJager
! de Jager, C., Nieuwenhuijzen, H., & van der Hucht, K. A. 1988, A&AS, 72, 259.
!            real(dp), intent(out) :: w
!            real(dp) :: log10w
!            include 'formats'
!            log10w = 1.769d0*log10_cr(L1/Lsun) - 1.676d0*log10_cr(T1) - 8.158d0
!            w = exp10_cr(log10w) 
! write(*,1) 'de_Jager log10 wind', log10w
!end subroutine deJager


! ========================================
!   === MAGNETIC MASS-LOSS QUENCHING ===
!      === ROTATIONAL ENHANCEMENT === 
! ========================================

subroutine MAG_WIND(id, L_phot, M_phot, R_phot, T_phot, w, ierr)   
  type (star_info), pointer :: s
  integer ::  h1, he4, nz
  real(dp) :: X, Y, Z, surface_h1, surface_he4, T1, L1
  real(dp), intent(in) :: L_phot, M_phot, R_phot, T_phot ! photospheric values (cgs)
  integer, intent(in) :: id
  integer, intent(out) :: ierr
  real(dp), intent(out) :: w
  logical, parameter :: dbg = .false.
  real(dp) :: log10wvan
  include 'formats'

  call get_star_ptr(id,s,ierr)  ! pointer to get data

  call magnetic(id,ierr)  ! quenching routine to obtain f_B
  call gammae(id,ierr)    ! call for rotational boost on mass-loss rates 

  T1 = T_phot 
  L1 = L_phot

  nz = s% nz
  h1 = s% net_iso(ih1)
  he4 = s% net_iso(ihe4)

  surface_h1 = s% xa(h1,1)
  surface_he4 = s% xa(he4,1)

  X = surface_h1
  Y = surface_he4     
  Z = 1 - (X + Y) 
  
if (T1 .GE. 9.d3 .AND. X > 0.3d0) then 
 call Vink(id, L_phot, M_phot, R_phot, T_phot, w, ierr) 
 Mdot_orig = w_vink
end if
 
if (T1 > 9.d3 .AND. X < 0.3d0) then  
 call Graefener(id, L_phot, M_phot, R_phot, T_phot, w, ierr)   
 Mdot_orig = w_graefener 
end if  

if (T1 < 9.d3) then  
 ! van Loon et al. 2005, A&A, 438, 273
   log10wvan = -5.65d0 + 1.05*log10_cr(L1/(1d4*Lsun)) - 6.3d0*log10_cr(T1/35d2)
   w_vanloon = exp10_cr(log10wvan)
   Mdot_orig = w_vanloon
end if  

 ! ========================================
 ! === Final mass-loss rate of the star ===
 ! ========================================

if (f_B > 0.d0 .AND. f_rot > 0.d0) then 
  w = Mdot_orig * f_B * f_rot  ! scaled by mass-loss quenching and rotational enhancement
 else 
  w = Mdot_orig 
  write(*,*) 'Either f_B or f_rot is zero.' 
  write(*,*) f_B
  write(*,*) f_rot
end if

  if (dbg) write(*,*) 'wind', w

end subroutine MAG_WIND

! =====================



subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)

   ! MESA provides routines for taking logarithms which will handle
   ! cases where you pass them zero, etc.  Take advantage of this!
   ! use crlibm_lib, only: safe_log10_cr
    use const_def, only: dp
    use star_def, only: maxlen_history_column_name

   integer, intent(in) :: id, id_extra, n
   character (len=maxlen_history_column_name) :: names(n)
   real(dp) :: vals(n)
   integer, intent(out) :: ierr
   type (star_info), pointer :: s
    
   ierr = 0
   call star_ptr(id, s, ierr)
   if (ierr /= 0) return
   ! note: do NOT add these names to history_columns.list
   ! the history_columns.list is only for the built-in log column options.
   ! it must not include the new column names you are adding here.

   ! extra column 1
   names(1) = "Bp"   ! polar magnetic field strength in gauss
   vals(1) = Bp

   ! extra column 2
   names(2) = "vesc" ! escape velocity in cgs
   vals(2) = vesc

   ! extra column 3
   names(3) = "vinf" ! terminal wind velocity in cgs
   vals(3) = vinf

   ! extra column 4
   names(4) = "etastar" ! equatorial magnetic confinement parameter
   vals(4) = etastar

   ! extra column 5
   names(5) = "Ra"     ! Alfven radius / stellar radius 
   vals(5) = Ra

   ! extra column 6
   names(6) = "Rc"     ! closure radius / stellar radius 
   vals(6) = Rc

   ! extra column 7
   names(7) = "f_B"    ! escaping wind fraction (quenching parameter)
   vals(7) = f_B

   ! extra column 8
   names(8) = "f_rot"  ! rotational enhancement 
   vals(8) = f_rot

  ! extra column 9
   names(9) = "R_init"
   vals(9) = R_init

  ! extra column 10
   names(10) = "B0"
   vals(10) = B0

  ! extra column 11
   names(11) = "Jbrake"  ! dJ / dt from ud-Doula, Owocki, Townsend
   vals(11) = Jbrake

  ! extra column 12
   names(12) = "Mdot_orig"  ! mass-loss rates before applying f_B and f_rot
   vals(12) = Mdot_orig

  ! extra column 13
   names(13) = "w_vink"
   vals(13) = w_vink

  ! extra column 14
   names(14) = "w_vanloon"
   vals(14) = w_vanloon

  ! extra column 15
   names(15) = "J_surf"     ! surface angular momentum reservior in cgs
   vals(15) = J_surf

!
! If you add / remove columns, it has to be synchronized with history column number 
! at the beginning of this file. 
!
   ierr = 0
  end subroutine data_for_extra_history_columns



  subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)
    use star_def, only: star_info, maxlen_profile_column_name
    use const_def, only: dp
    integer, intent(in) :: id, id_extra, n, nz
    character (len=maxlen_profile_column_name) :: names(n)
    real(dp) :: vals(nz,n)
    integer, intent(out) :: ierr
    type (star_info), pointer :: s
    integer :: k
    ierr = 0
    call star_ptr(id, s, ierr)
    if (ierr /= 0) return
    
    !note: do NOT add the extra names to profile_columns.list
    ! the profile_columns.list is only for the built-in profile column options.
    ! it must not include the new column names you are adding here.
    
    ! here is an example for adding a profile column
    !if (n /= 1) stop 'data_for_extra_profile_columns'
    
    names(1) = 'extra_jdot'
    do k = 1, nz
       vals(k,1) = s%extra_jdot(k)  ! angular momentum removed in a given layer.
    end do
         
  end subroutine data_for_extra_profile_columns


end module run_star_extras

